package algorithms;


import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;
import org.apache.bcel.util.Class2HTML;


/**
@ Class: InsertDeadCodeMethods
@ Author: Suma Venkatesh
@ Description:   This algorithm is one kind of Smoke and Mirror technique which inserts irrelevant code (dead code)
  into methods that will never be executed
*/

public class InsertDeadCodeMethods {



	public boolean apply(String classname) {
		ClassGen cgen = null;
		boolean sucess = false;
	    try {
		   JavaClass clazz = Repository.lookupClass("examples/" + classname);
		   cgen = new ClassGen(clazz);
		   cgen = appendDeadCodeMethods(classname,clazz,cgen);
		   sucess = true;
	     }
	     catch(Exception e) {System.out.println("Exception " +e ); return(sucess);}
		return(sucess);
     }

    public static ClassGen appendDeadCodeMethods(String classname,JavaClass clazz,ClassGen cgen) {
         try {
			Method[] methods = clazz.getMethods();
			MethodGen mymgen = null;
			ConstantPoolGen cpg = new ConstantPoolGen(clazz.getConstantPool());
		   	for (int i=0; i < methods.length; i++) {
				mymgen = new MethodGen(methods[i], classname, cpg);
				if ((mymgen.isInterface() || mymgen.isAbstract())) {
		           // Nothing to do! (that was easy)
		           return(cgen);
      			}
		    }

            InstructionList il= mymgen.getInstructionList();
            Instruction[] ins = il.getInstructions();
			InstructionHandle[] ihs=il.getInstructionHandles();
			double rand_var;
			double rand_inst_add;
			rand_inst_add=Math.random()*10;

			for (int j=0; j<rand_inst_add; j++) {
			    rand_var=Math.random();
			    // code that randomly adds instructions
			    if (rand_var<.1) {
			         il.append(new GOTO(ihs[(int) ((9999*rand_var)%ihs.length)]));
			         il.append(ins[0]);

			    } else if (rand_var<.2) {
			        il.append(new BIPUSH((byte)(((int)(rand_var*9999))%255)));
			        il.append(ins[1]);

			    } else if (rand_var<.3) {
			    	   if (mymgen.getMaxLocals()>1) {
			              il.append(new RET(((int)((rand_var*9999))%(mymgen.getMaxLocals()-1))+1));
			              il.append(ins[2]);

			           }
			   } else if (rand_var<.4) {
			             il.append(new DUP());
			             il.append(ins[3]);

			   } else if (rand_var<.5) {
			           il.append(new SWAP());
			           il.append(ins[4]);

			   } else if (rand_var<.6) {
			          if (mymgen.getMaxLocals()>1) {
			               il.append(new ILOAD(((int)((rand_var*9999))%(mymgen.getMaxLocals()-1))+1));
			               il.append(ins[5]);

			          }
			   } else if (rand_var<.7) {
			           il.append(new IMUL());
			           il.append(ins[6]);

			   } else if (rand_var<.8) {
			          if (mymgen.getMaxLocals()>1) {
			              il.append(new ISTORE(((int)((rand_var*9999))%(mymgen.getMaxLocals()-1))+1));
			              il.append(ins[7]);

			          }
			   } else if (rand_var<.9) {
			           il.append(new IADD());
			           il.append(ins[8]);

			   }
			 }

			 // All done!   Put it back together...
			 mymgen.setInstructionList(il);
			 methods[0]=mymgen.getMethod();
			 cgen.setMethods(methods);
			 cgen.setConstantPool(cpg);
			 clazz.setConstantPool(cpg.getFinalConstantPool());
			 clazz.dump(new java.io.File("C:/JHide/Obfuscated_Classes/",clazz.getClassName().replace('.', java.io.File.separatorChar) + "_obf" +".class"));
			 convertToHTML(clazz,"Obf");
  }


  catch(Exception e) { System.out.println("Exception " +e );return(cgen);}
  return(cgen) ;

}

 public static void convertToHTML(JavaClass clazz,String type)   throws java.io.IOException {
    Class2HTML conObj = new Class2HTML(clazz,"C:/JHide/Obfuscated_Classes/HTMLFiles/" + type);
 }

}