package algorithms.Method2RMadness;

/** Merges all static methods with the same signature into one method.
 *
 * <p>
 * Used by Method2RMadness obfuscator.
 *
 * @author Kelly Heffner (<a href="mailto:kheffner@cs.arizona.edu">kheffner@cs.arizona.edu</a>)
 * @version 1.0, May 8, 2002
 */

public class MethodMerger {

    private java.util.HashMap myMergeSets = new java.util.HashMap();
    private java.util.HashMap myMergedMethods = new java.util.HashMap();

    private utils.ConfigProperties myProps;


    /**
       Applys the method merging.
    */
    public void apply(utils.ClassFileCollection cfc)
        throws Exception
    {

        try{

            java.util.ArrayList allClasses = new java.util.ArrayList();

            java.util.Iterator fileItr = cfc.classes();
            while(fileItr.hasNext()){
                String className = (String)fileItr.next();

                if(className.indexOf(".class") >= 0)
                    className = className.substring(0, className.indexOf(".class"));

                //first, we open this class, and build the sets of mergable methods
                de.fub.bytecode.classfile.JavaClass mergeTarget = cfc.getClass(className);

                java.util.HashMap oneClassSets = buildMergeSets(mergeTarget);

                //now for each set, merge them!!!
                java.util.Iterator setItr = oneClassSets.values().iterator();
                while(setItr.hasNext()){
                   MergeInfo toMerge =(MergeInfo)setItr.next();

                    mergeTarget = merge(toMerge, allClasses, mergeTarget);
                }

                allClasses.add(mergeTarget);
                //                cfc.addClass(mergeTarget);
            }

            //now that all classes have been added, patch method calls

            fixMethodCalls(allClasses, cfc);

        }
        catch(Exception e){
            e.printStackTrace();
        }
    }


    private de.fub.bytecode.classfile.JavaClass merge(MergeInfo mergeInfo,
                                                      java.util.ArrayList allClasses,
                                                      de.fub.bytecode.classfile.JavaClass mergeTarget)
        throws Exception
    {
        java.util.ArrayList mergeSet = mergeInfo.mergeSet;
        int accessFlags = 0;
        de.fub.bytecode.generic.Type returnType = null;

        de.fub.bytecode.generic.ClassGen targetGen =
            new de.fub.bytecode.generic.ClassGen(mergeTarget);

        de.fub.bytecode.generic.ConstantPoolGen cpg = targetGen.getConstantPool();

        de.fub.bytecode.generic.Type[] argTypes = null;

        String methodName = getUnusedName(mergeTarget);

        de.fub.bytecode.generic.InstructionFactory factory =
            new de.fub.bytecode.generic.InstructionFactory(cpg);

        java.util.ArrayList exceptions = new java.util.ArrayList();
        java.util.ArrayList instructions = new java.util.ArrayList();
        de.fub.bytecode.generic.InstructionHandle [] handleKeys =
            new de.fub.bytecode.generic.InstructionHandle[mergeSet.size()];
        int flagIndex = 0;

        //int offset = 0;

        de.fub.bytecode.generic.InstructionList newInstList =
            new de.fub.bytecode.generic.InstructionList();

        //pull the methods from the list to get some info :)
        for(int i = 0; i < mergeSet.size(); i++){
            de.fub.bytecode.classfile.Method repMeth =
                (de.fub.bytecode.classfile.Method)mergeSet.get(i);

            de.fub.bytecode.generic.MethodGen repGen =
                new de.fub.bytecode.generic.MethodGen(repMeth,
                                                      mergeTarget.getClassName(),
                                                      cpg);

            accessFlags |= repMeth.getAccessFlags();
            returnType = repGen.getReturnType();
            argTypes = repGen.getArgumentTypes();

            //append the code
            //first, append the list of instructions from this
            //method
            de.fub.bytecode.generic.InstructionList oneMethodCode =
                repGen.getInstructionList();

            handleKeys[i] = newInstList.append(oneMethodCode);

            flagIndex = 0;
            for(int k = 0; k < argTypes.length; k++){

                flagIndex += argTypes[k].getSize();
            }


            de.fub.bytecode.generic.CodeExceptionGen [] handlers =
                repGen.getExceptionHandlers();
            for(int e = 0; e < handlers.length; e++)
                exceptions.add(handlers[e]);

        }

        de.fub.bytecode.generic.Type[] newArgs =
            new de.fub.bytecode.generic.Type[argTypes.length+1];

        String [] argNames = new String[newArgs.length];
        for(int i = 0; i < argTypes.length; i++){
            argNames[i] = "keldaANDmartio" + i;
            newArgs[i] = argTypes[i];
        }
        argNames[argTypes.length] = "switchVar";
        newArgs[argTypes.length] = de.fub.bytecode.generic.Type.INT;


        instructions.clear();
        newInstList.setPositions();
        for(int patch = 0; patch < handleKeys.length; patch++){

            de.fub.bytecode.generic.InstructionHandle ih = handleKeys[patch];
            //next, append a jump to this instuction set for the corresponding
            //int --BACKWARDS
            //create an equals branch to ih

            instructions.add(new de.fub.bytecode.generic.IF_ICMPEQ(ih));
            instructions.add(new de.fub.bytecode.generic.ILOAD(flagIndex));
            instructions.add(new de.fub.bytecode.generic.BIPUSH((byte)patch));
        }

        java.util.Iterator instItr = instructions.iterator();
        while(instItr.hasNext()){
            de.fub.bytecode.generic.Instruction currInst =
                (de.fub.bytecode.generic.Instruction)instItr.next();
            if(currInst instanceof de.fub.bytecode.generic.BranchInstruction)
                newInstList.insert((de.fub.bytecode.generic.BranchInstruction)currInst);
            else
                newInstList.insert(currInst);
        }

        de.fub.bytecode.generic.MethodGen newMethod =
            new de.fub.bytecode.generic.MethodGen(accessFlags, returnType, newArgs,
                                                  argNames, methodName,
                                                  mergeTarget.getClassName(),
                                                  newInstList, cpg);

        trimAccessFlags(newMethod);
        newMethod.setMaxStack();
        newMethod.setMaxLocals();

        for(int e = 0; e < exceptions.size(); e++){
            de.fub.bytecode.generic.CodeExceptionGen ex =
                (de.fub.bytecode.generic.CodeExceptionGen)exceptions.get(e);

            newMethod.addExceptionHandler(ex.getStartPC(), ex.getEndPC(),
                                          ex.getHandlerPC(), ex.getCatchType());
        }

        mergeInfo.className = newMethod.getClassName();
        mergeInfo.methodName = newMethod.getName();
        mergeInfo.signature = newMethod.getSignature();
        mergeInfo.retType = newMethod.getReturnType();
        mergeInfo.argTypes = newMethod.getArgumentTypes();


        for(int m = 0; m < mergeSet.size(); m++)
            targetGen.removeMethod((de.fub.bytecode.classfile.Method)mergeSet.get(m));

        targetGen.addMethod(newMethod.getMethod());

        de.fub.bytecode.classfile.JavaClass newClass =
            targetGen.getJavaClass();

        de.fub.bytecode.classfile.ConstantPool pool =
            cpg.getFinalConstantPool();

        newClass.setConstantPool(pool);

        return newClass;
    }

    private void fixMethodCalls(java.util.ArrayList allClasses, utils.ClassFileCollection cfc) throws Exception
    {

        java.util.Iterator itr = allClasses.iterator();
        while(itr.hasNext()){

            de.fub.bytecode.classfile.JavaClass currClass =
                (de.fub.bytecode.classfile.JavaClass)itr.next();

            de.fub.bytecode.classfile.Method [] methods = currClass.getMethods();

            de.fub.bytecode.generic.ClassGen currCGen =
                new de.fub.bytecode.generic.ClassGen(currClass);

            //currCGen.setMethods(new de.fub.bytecode.classfile.Method[]{});

            de.fub.bytecode.generic.ConstantPoolGen cpg = currCGen.getConstantPool();

            de.fub.bytecode.generic.InstructionFactory factory =
                new de.fub.bytecode.generic.InstructionFactory(cpg);

            for(int i = 0; i < methods.length; i++){
                de.fub.bytecode.classfile.Method currMeth = methods[i];
                de.fub.bytecode.generic.MethodGen currGen =
                    new de.fub.bytecode.generic.MethodGen(currMeth,
                                                          currClass.getClassName(),
                                                          cpg);

                de.fub.bytecode.generic.InstructionList il =
                    currGen.getInstructionList();

                if(il != null){

                    for(de.fub.bytecode.generic.InstructionHandle ih = il.getStart();
                        ih != null ; ih = ih.getNext()){

                        de.fub.bytecode.generic.Instruction inst= ih.getInstruction();
                        if(inst instanceof de.fub.bytecode.generic.INVOKESTATIC){
                            de.fub.bytecode.generic.INVOKESTATIC invoke =
                                (de.fub.bytecode.generic.INVOKESTATIC)inst;

                            java.util.Iterator mergeSetItr = myMergeSets.values().iterator();
                            while(mergeSetItr.hasNext()){
                                MergeInfo info =
                                    ((MergeInfo)mergeSetItr.next());

                                java.util.ArrayList mergeSet = info.mergeSet;


                                String className = info.className;
                                for(int p = 0; p < mergeSet.size(); p++){

                                    de.fub.bytecode.classfile.Method possible =
                                        (de.fub.bytecode.classfile.Method)mergeSet.get(p);


                                    if(invoke.getClassName(cpg).equals(info.className)
                                       &&
                                       invoke.getMethodName(cpg).equals(possible.getName())
                                       &&
                                       invoke.getSignature(cpg).equals(possible.getSignature())){

                                        de.fub.bytecode.generic.Instruction newinst =
                                            new de.fub.bytecode.generic.BIPUSH((byte)p);
                                        ih.setInstruction(newinst);

                                        inst = factory.createInvoke(info.className,
                                                                    info.methodName,
                                                                    info.retType,
                                                                    info.argTypes,
                                                                    de.fub.bytecode.Constants.INVOKESTATIC);

                                        il.append(ih, inst);
                                        ih = ih.getNext();
                                    }
                                }
                            }
                        }
                    }
                }
                currGen.setInstructionList(il);
                currGen.setMaxStack();
                currGen.setMaxLocals();
                //currCGen.addMethod(currGen.getMethod());
                currCGen.replaceMethod(currMeth, currGen.getMethod());
            }
            de.fub.bytecode.classfile.JavaClass newClass =
                currCGen.getJavaClass();

            de.fub.bytecode.classfile.ConstantPool pool =
                cpg.getFinalConstantPool();

            newClass.setConstantPool(pool);

            cfc.addClass(newClass);
        }
    }


    private String getUnusedName(de.fub.bytecode.classfile.JavaClass inFile)
    {
        java.util.HashMap map = new java.util.HashMap();
        de.fub.bytecode.classfile.Method [] methods = inFile.getMethods();
        for(int i = 0; i < methods.length; i++)
            map.put(methods[i].getName(), "name");

        int i = 0;
        String proposed = "keldaANDmartio";
        while(map.containsKey(proposed+i))
            i++;

        return proposed+i;
    }

    private java.util.HashMap buildMergeSets(de.fub.bytecode.classfile.JavaClass mergeTarget)
    {

        de.fub.bytecode.classfile.Method[] methods = mergeTarget.getMethods();
        java.util.HashMap thisClassSets = new java.util.HashMap();

        for(int methI = 0; methI < methods.length; methI++){
            de.fub.bytecode.classfile.Method currMeth = methods[methI];

            String methSig = currMeth.getSignature();

            //at the moment, dont care about dynamic methods
            if(!currMeth.isStatic())
                continue;

            //also need to cut out any method that should not be messed with
            if(currMeth.getName().equals("main") ||
               currMeth.getName().equals("<clinit>") ||
               currMeth.isSynchronized() || currMeth.isVolatile() || currMeth.isNative())
                continue;


            //now the fun part, add this method into the ArrayList that hashes
            //to its signature

            String key = mergeTarget.getClassName() + methSig;

            MergeInfo info =
                (MergeInfo)myMergeSets.get(key);

            if(info == null){
                java.util.ArrayList list = new java.util.ArrayList();
                info = new MergeInfo(list);

                myMergeSets.put(key, info);
                thisClassSets.put(key, info);
            }

            info.mergeSet.add(currMeth);
        }
        System.out.println("CLASS SETS: " + thisClassSets);
        return thisClassSets;

    }




    //removes clashing flags
    private static void trimAccessFlags(de.fub.bytecode.generic.MethodGen acc)
    {
        if(acc.isProtected()){
            if(acc.isPublic()){
                acc.isProtected(false);
            }
            acc.isPrivate(false);
        }
    }

    public static void main(String [] args) throws java.lang.Exception
    {
       utils.ClassFileCollection cfc =  new utils.ClassFileCollection(args[0]);
       MethodMerger merger =   new MethodMerger();
       merger.apply(cfc);
       cfc.saveJar("MERGED.jar");
    }
}
