package algorithms.Method2RMadness;


import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;



/** Makes all fields and methods in an application public.
  *
  * <p>
  * Used by Method2RMadness obfuscator.
  *
  * @author Martin Stepp (<a href="mailto:stepp@cs.arizona.edu">stepp@cs.arizona.edu</a>)
  * @version 1.0, May 8, 2002
  */
public class Publicizer implements de.fub.bytecode.Constants
{

	/* Turn to true to print debugging messages. */
	private static final boolean DEBUG = false;

	/* Required for SandMark's interface. */
	private utils.ConfigProperties myProps = new utils.ConfigProperties();


	/* Runs a quick test of this obfuscator. */
	public static void main(String[] args) throws Exception
	{
		if (args.length < 1)
		{
			System.out.println("Usage: Publicizer JAR_FILE");
			System.exit(1);
		}

		try
		{
			utils.ClassFileCollection cfc = new utils.ClassFileCollection(args[0]);
			Publicizer obfuscator = new Publicizer();
			obfuscator.apply(cfc);
			// printCFC(cfc);
			cfc.saveJar("HARDCODE.jar");
		}
		catch (java.io.IOException e)
		{
			System.err.println("I/O Error: " + e.getMessage());
		}
	}


	/** Prints the contents of the given class file collection to System.out. */
	private static void printCFC(utils.ClassFileCollection cfc)
		throws Exception
	{
		java.util.Iterator itr = cfc.classes();
		while (itr.hasNext())
		{
			String className = (String)itr.next();
			System.out.println("Class " + className);
			de.fub.bytecode.classfile.JavaClass jc = cfc.getClass(className);

			de.fub.bytecode.generic.ClassGen cgen = new de.fub.bytecode.generic.ClassGen(jc);
			de.fub.bytecode.generic.ConstantPoolGen cpgen = cgen.getConstantPool();

			Field[] fields = cgen.getFields();
			int numFields = fields.length;
			for (int fieldNum = 0;  fieldNum < numFields;  fieldNum++)
				System.out.println(fields[fieldNum]);

			// print all methods in this class
			int numMethods = cgen.getMethods().length;
			for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
			{
				// reorder each method
				de.fub.bytecode.classfile.Method meth = cgen.getMethodAt(methodNum);
				de.fub.bytecode.generic.MethodGen methGen  = new de.fub.bytecode.generic.MethodGen(meth, className, cpgen);
				System.out.println(meth + "{");
				System.out.println(methGen.getInstructionList() + "\n");
				System.out.println("}");
			}
		}
	}


	/** Makes public the fields/methods in this ClassFileCollection's classes. */
	public void apply(utils.ClassFileCollection cfc)
		throws Exception
	{
		// create a class hierarchy for later seeing if methods are renamable
           //CC		sandmark.util.ClassHierarchy ch = new sandmark.util.ClassHierarchy(cfc, myProps);

		java.util.ArrayList classNameList = new java.util.ArrayList();
		java.util.Iterator itr = cfc.classes();
		while (itr.hasNext())
			classNameList.add((String)itr.next());

		int numClasses = classNameList.size();
		for (int classNum = 0;  classNum < numClasses;  classNum++)
		{
			String className = (String)classNameList.get(classNum);
			de.fub.bytecode.classfile.JavaClass jc = cfc.getClass(className);

			de.fub.bytecode.generic.ClassGen cgen = new de.fub.bytecode.generic.ClassGen(jc);
			de.fub.bytecode.generic.ConstantPoolGen cpgen = cgen.getConstantPool();

			// publicize this class
			cgen.isPublic(true);
			cgen.isPrivate(false);
			cgen.isProtected(false);

			// publicize all methods in this class
			int numMethods = cgen.getMethods().length;
			for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
			{
				// reorder each method
				de.fub.bytecode.classfile.Method meth = cgen.getMethodAt(methodNum);
				de.fub.bytecode.generic.MethodGen methGen  = new de.fub.bytecode.generic.MethodGen(meth, className, cpgen);

				methGen.isPublic(true);
				methGen.isPrivate(false);
				methGen.isProtected(false);
				de.fub.bytecode.classfile.Method newMeth = methGen.getMethod();

				// replace method in class
				cgen.replaceMethod(meth, newMeth);
			}

			Field[] fields = cgen.getFields();
			int numFields = fields.length;
			for (int fieldNum = 0;  fieldNum < numFields;  fieldNum++)
			{
				FieldGen fgen = new FieldGen(fields[fieldNum], cpgen);
				fgen.isPublic(true);
				fgen.isPrivate(false);
				fgen.isProtected(false);
				Field newField = fgen.getField();
				cgen.replaceField(fields[fieldNum], newField);
			}

			// done replacing all methods/fields in this class;
			// now replace old class with new one
			cfc.addClass(cgen.getJavaClass());
		}
	}


	

	
}
