package algorithms.Method2RMadness;

/** Splits each nonstatic method of the class into a nonstatic method
  * with the same name and signature as before, and a static method
  * that contains the method body of the original method.
  *
  * <p> Used by Method2RMadness obfuscator.
  *
  * @author Kelly Heffner (<a href="mailto:kheffner@cs.arizona.edu">kheffner@cs.arizona.edu</a>)
  * @version 1.0, May 8, 2002
  * Picked this algorithm from SandMark
  */
public class StaticSplit {



        //A list of names in this class that have already been used, to keep a fast way to
        //determine if the name I want to use is taken.
        private java.util.HashMap myUsedNames;
        private utils.ConfigProperties myProps;

        public static boolean DEBUG = false;

        public static void main(String [] args) throws java.lang.Exception
        {
                

                utils.ClassFileCollection cfc =
                        new utils.ClassFileCollection(args[0]);

               StaticSplit splitter =  new StaticSplit();

                java.util.Iterator fileItr = cfc.classes();
                splitter.apply(cfc, (String)fileItr.next());
                cfc.saveJar("SPLIT.jar");
        }

        /**
                Modifies the given class file collection, class, and method so that
                the each method of the class is split into two parts.  The first part is
                a nonstatic method with the same name and signature as the specified
                method.  The second part is a static method with a unique name within
                the class that contains the method body of the original method.  If this
                method is passed information about a static method, the method will not
                be modified.
                @param cfc the collection of java classes dependent on this class
                @param classname the name of the class which contains the methods to change
        */
        public void apply(utils.ClassFileCollection cfc, java.lang.String classname)
                throws java.io.IOException
        {

            try{

                myUsedNames = new java.util.HashMap();

                if(DEBUG) System.out.println("looking for class named: " + classname);
            
                de.fub.bytecode.classfile.JavaClass methodHome = cfc.getClass(classname /*+".class"*/);
                if(methodHome.isInterface())
                        return;

                de.fub.bytecode.generic.ClassGen classEdit = new de.fub.bytecode.generic.ClassGen(methodHome);


                de.fub.bytecode.generic.ConstantPoolGen cpEdit = classEdit.getConstantPool();

                //get a method and methodGen for the method to split
                de.fub.bytecode.classfile.Method [] allMethods = methodHome.getMethods();

                for(int i = 0; i < allMethods.length; i++)
                {
                        myUsedNames.put(allMethods[i].getName()+allMethods[i].getName(), Boolean.TRUE);
                }

                for(int i = 0; i < allMethods.length; i++)
                {
                    if(!allMethods[i].getName().equals("<init>"))
                    {
                        de.fub.bytecode.generic.MethodGen temp =
                                new de.fub.bytecode.generic.MethodGen(allMethods[i], classname, cpEdit);
                        if(!temp.isStatic() && !temp.isAbstract() && !temp.isNative() && !temp.isSynchronized() && !temp.isVolatile())
                        {
                                classEdit.removeMethod(allMethods[i]);
                                split(temp, classEdit, cpEdit);
                        }
                    }
                }
                de.fub.bytecode.classfile.JavaClass newVersion = classEdit.getJavaClass();
                cfc.addClass(newVersion);

                //update the cfc
                //cfc.addClass

                //cfc.saveJar("hardcode.jar");
                //cfc.close();
            }catch(Exception e)
            {
                System.out.println(e);
                e.printStackTrace();
            }

        }


        public void split(de.fub.bytecode.generic.MethodGen origMethEdit,
                                          de.fub.bytecode.generic.ClassGen classEdit,
                                          de.fub.bytecode.generic.ConstantPoolGen cpEdit)
        {
            try{
                if(DEBUG)System.out.println("Method being edited: " + origMethEdit.getName());

                //rip out the code in the method, and wrap it in a static method named methname$args$1 or 2 or 3...
                //get list of code
                de.fub.bytecode.generic.InstructionList realCode = origMethEdit.getInstructionList();

                //find a suitable name for our new method - our first attempt will be originalmethodname
                //concatonated with its arg types
                de.fub.bytecode.generic.Type [] origArgTypes = origMethEdit.getArgumentTypes();
                if(DEBUG)System.out.println(origArgTypes);

                //now append the reference to "this" to the front
                de.fub.bytecode.generic.Type [] argTypes = new de.fub.bytecode.generic.Type[origArgTypes.length+1];

                //need to figure out type of this -- de.fub.bytecode.generic.Type.getType(/*signature*/);
                //de.fub.bytecode.generic.Type thisType =
                de.fub.bytecode.classfile.ConstantClass cc =
                    (de.fub.bytecode.classfile.ConstantClass)cpEdit.getConstant(classEdit.getClassNameIndex());
                if(DEBUG)System.out.println( "Constant Class: " + cc);
                de.fub.bytecode.classfile.ConstantUtf8 thisName =
                    (de.fub.bytecode.classfile.ConstantUtf8)cpEdit.getConstant(cc.getNameIndex());
                argTypes[0] = de.fub.bytecode.generic.Type.getType("L" + thisName.getBytes() + ";");

                for(int i = 0; i < origArgTypes.length; i++)
                        argTypes[i+1] = origArgTypes[i];


                String[] argNames = new String[origArgTypes.length+1];
                argNames[0] = "martio";
                for(int i = 0; i < argNames.length; i++)
                        argNames[i] = "arg"+i;

                String newName = origMethEdit.getName();
                for(int i = 1; i < argTypes.length; i++)
                {
                        newName += (!(argTypes[i] instanceof de.fub.bytecode.generic.ReferenceType))?argTypes[i].toString():"ref"; //this may be a problem with parsing
                }
                if(argTypes.length == 1)
                        newName += "PARENPAREN";

                int number = 0;
                while(myUsedNames.containsKey(newName+number +origMethEdit.getSignature()))
                {
                        number++;
                }
                newName = newName+number;
                myUsedNames.put(newName+number +origMethEdit.getSignature(), Boolean.TRUE);

                //now create a new method called methname$args$number, with the same access as its parent + static
                de.fub.bytecode.generic.MethodGen newMethEdit = new de.fub.bytecode.generic.MethodGen(
                                origMethEdit.getAccessFlags(), origMethEdit.getReturnType(),
                                argTypes, argNames,     newName, origMethEdit.getClassName(),
                                realCode, cpEdit);

                newMethEdit.isStatic(true);
                newMethEdit.setMaxLocals(origMethEdit.getMaxLocals());
                newMethEdit.setMaxStack(origMethEdit.getMaxStack());

                //need to add in the exception handlers from the old method
                de.fub.bytecode.generic.CodeExceptionGen [] handlers =
                        origMethEdit.getExceptionHandlers();
                for(int i = 0; i < handlers.length; i++){
                        de.fub.bytecode.generic.CodeExceptionGen handler = handlers[i];
                        newMethEdit.addExceptionHandler(handler.getStartPC(),
                                handler.getEndPC(), handler.getHandlerPC(), handler.getCatchType());
                }

                //that should be it for this new methods construction -> add it in!
                classEdit.addMethod(newMethEdit.getMethod());

                //change the body of the original to call this new static method
                //this new body should push all of its params, and invokestatic
                de.fub.bytecode.generic.InstructionList newBody = new de.fub.bytecode.generic.InstructionList();
                de.fub.bytecode.generic.InstructionFactory codeFactory =
                        new de.fub.bytecode.generic.InstructionFactory(classEdit, cpEdit);
                int offset = 0;
                for(int i = 0; i < argTypes.length; i++)
                {
                        newBody.append(codeFactory.createLoad(argTypes[i],offset));
                        offset += argTypes[i].getSize();
                }
                newBody.append(codeFactory.createInvoke(origMethEdit.getClassName(),
                                                        newName, origMethEdit.getReturnType(),
                                                        argTypes, de.fub.bytecode.Constants.INVOKESTATIC));

                newBody.append(codeFactory.createReturn(origMethEdit.getReturnType()));


                String [] oldArgs = new String [argTypes.length-1];
                for(int i = 0; i < oldArgs.length; i++)
                    {
                        oldArgs[i] = "arg" + i;
                    }

                de.fub.bytecode.generic.MethodGen origMethEdit2 = new de.fub.bytecode.generic.MethodGen(
                                origMethEdit.getAccessFlags(), origMethEdit.getReturnType(),
                                origMethEdit.getArgumentTypes(), oldArgs ,
                                origMethEdit.getName(), origMethEdit.getClassName(),
                                newBody, cpEdit);

                origMethEdit2.setMaxStack();

                classEdit.addMethod(origMethEdit2.getMethod());
            }catch(Exception e)
                {
                    e.printStackTrace();
                }
        }


        


}
