package algorithms;

import org.apache.bcel.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.Class2HTML;
import java.util.*;

/**
@ Class: MethodsMerging
@ Author: Suma Venkatesh
@ Description:   This algorithm include merging all possible program constructs.
  A method merging is a transformation, where logically dependent methods are merged.
*/

public class MethodsMerging {

	private java.util.HashMap myMergeSets = new java.util.HashMap();
    private java.util.HashMap myMergedMethods = new java.util.HashMap();

	public boolean apply(String classname ) {
		ClassGen cgen = null;
		boolean sucess = false;
		try {
			JavaClass clazz = Repository.lookupClass("examples/" + classname);
		    cgen = new ClassGen(clazz);
			mergeMethods(clazz);
			sucess = true;

		}
	    catch(Exception e) {System.out.println("Exception" +e.getMessage()+e.fillInStackTrace()); return(sucess);}
		return(sucess);
	}

	public void mergeMethods(JavaClass clazz)throws Exception {

		try {
			//first, we open this class, and build the sets of mergable methods
			java.util.HashMap oneClassSets = buildMergeSets(clazz);

			//now for each set, merge them
			java.util.Iterator setItr = oneClassSets.values().iterator();
			while(setItr.hasNext()){
			      MergeInfo toMerge =(MergeInfo)setItr.next();
				  System.out.println("toMerge.toString()" + toMerge.toString());
			      clazz = merge(toMerge, clazz);
			      fixMethodCalls(clazz);

            }

		 } catch(Exception e) { System.out.println("Exception" +e.getMessage() + e.fillInStackTrace()); }

	 }//MethodMerging

	 private java.util.HashMap buildMergeSets(JavaClass mergeTarget){
			Method[] methods = mergeTarget.getMethods();
	        java.util.HashMap thisClassSets = new java.util.HashMap();

	        for(int methI = 0; methI < methods.length; methI++){
	             Method currMeth = methods[methI];
	             String methSig = currMeth.getSignature();
				 System.out.println("methSig -- currMeth.getName() " + methSig + currMeth.getName());

	        	 //do not merge the following methods
	             if(currMeth.getName().equals("main") ||
	                currMeth.getName().equals("<clinit>") ||
	                currMeth.getName().equals("<init>") ||
	                currMeth.isSynchronized() || currMeth.isVolatile() || currMeth.isNative())
	                continue;


	             //Add the method into the ArrayList

	             String key = mergeTarget.getClassName() + methSig;
				 System.out.println("key"+key);
	             MergeInfo info = (MergeInfo)myMergeSets.get(key);
	             if(info == null){
	                 ArrayList list = new ArrayList();
	                 info = new MergeInfo(list);

	                 myMergeSets.put(key, info);
	                 thisClassSets.put(key, info);
	             }

	             info.mergeSet.add(currMeth);

	         }
	         System.out.println("CLASS SETS: " + thisClassSets );
	         return thisClassSets;

    }

	private JavaClass merge(MergeInfo mergeInfo,JavaClass mergeTarget) throws Exception {
        ArrayList mergeSet = mergeInfo.mergeSet;
        int accessFlags = 0;
        Type returnType = null;
        ClassGen targetGen = new ClassGen(mergeTarget);
		ConstantPoolGen cpg = targetGen.getConstantPool();
       	Type[] argTypes = null;
        String methodName = getUnusedName(mergeTarget);
        InstructionFactory factory = new InstructionFactory(cpg);
        ArrayList exceptions = new ArrayList();
        ArrayList instructions = new ArrayList();
        InstructionHandle [] handleKeys = new InstructionHandle[mergeSet.size()];
        int flagIndex = 0;


        InstructionList newInstList =  new InstructionList();

        //pull the methods from the list to get some info
        for(int i = 0; i < mergeSet.size(); i++){
            Method repMeth = (Method)mergeSet.get(i);
            MethodGen repGen = new MethodGen(repMeth, mergeTarget.getClassName(),cpg);
            accessFlags |= repMeth.getAccessFlags();
            returnType = repGen.getReturnType();
            argTypes = repGen.getArgumentTypes();

            //append the code, append the list of instructions from this method
            InstructionList oneMethodCode = repGen.getInstructionList();
            handleKeys[i] = newInstList.append(oneMethodCode);

            flagIndex = 0;
            for(int k = 0; k < argTypes.length; k++){

                flagIndex += argTypes[k].getSize();
            }


            CodeExceptionGen [] handlers = repGen.getExceptionHandlers();
            for(int e = 0; e < handlers.length; e++)
                exceptions.add(handlers[e]);

        }

        Type[] newArgs = new Type[argTypes.length+1];

        String [] argNames = new String[newArgs.length];
        for(int i = 0; i < argTypes.length; i++){
            argNames[i] = "A+MERGE+B" + i;
            newArgs[i] = argTypes[i];
        }
        argNames[argTypes.length] = "mergedVar";
        newArgs[argTypes.length] = Type.INT;


        instructions.clear();
        newInstList.setPositions();
        for(int patch = 0; patch < handleKeys.length; patch++){

            InstructionHandle ih = handleKeys[patch];
            //next, append a jump to this instuction set
            instructions.add(new IF_ICMPEQ(ih));
            instructions.add(new ILOAD(flagIndex));
            instructions.add(new BIPUSH((byte)patch));
        }

        Iterator instItr = instructions.iterator();
        while(instItr.hasNext()){
            Instruction currInst = (Instruction)instItr.next();
            if(currInst instanceof BranchInstruction)
                newInstList.insert((BranchInstruction)currInst);
            else
                newInstList.insert(currInst);
        }

        MethodGen newMethod =  new MethodGen(accessFlags, returnType, newArgs,argNames, methodName,mergeTarget.getClassName(),newInstList, cpg);
        trimAccessFlags(newMethod);
        newMethod.setMaxStack();
        newMethod.setMaxLocals();

        for(int e = 0; e < exceptions.size(); e++){
            CodeExceptionGen ex = (CodeExceptionGen)exceptions.get(e);
            newMethod.addExceptionHandler(ex.getStartPC(), ex.getEndPC(), ex.getHandlerPC(), ex.getCatchType());
        }

        mergeInfo.className = newMethod.getClassName();
        mergeInfo.methodName = newMethod.getName();
        mergeInfo.signature = newMethod.getSignature();
        mergeInfo.retType = newMethod.getReturnType();
        mergeInfo.argTypes = newMethod.getArgumentTypes();


        for(int m = 0; m < mergeSet.size(); m++)
    	     targetGen.removeMethod((Method)mergeSet.get(m));
		targetGen.addMethod(newMethod.getMethod());
	    JavaClass newClass = targetGen.getJavaClass();
	    ConstantPool pool = cpg.getFinalConstantPool();
	    newClass.setConstantPool(pool);
	    return newClass;
    }
     private String getUnusedName(JavaClass inFile) {
	        HashMap map = new HashMap();
	        Method [] methods = inFile.getMethods();
	        for(int i = 0; i < methods.length; i++)
	            map.put(methods[i].getName(), "name");

	        int i = 0;
	        String proposed = "A+MERGE+B";
	        while(map.containsKey(proposed+i))
	            i++;

	        return proposed+i;
	 }
	  //removes clashing flags
	 private static void trimAccessFlags(MethodGen acc) {
	       if(acc.isProtected()){
	          if(acc.isPublic()){
	             acc.isProtected(false);
	          }
	          acc.isPrivate(false);
	       }
    }

    private void fixMethodCalls(JavaClass clazz) throws Exception {


	      Method [] methods = clazz.getMethods();
		  ClassGen cgen = new ClassGen(clazz);
	      ConstantPoolGen cpg = cgen.getConstantPool();
	      InstructionFactory factory =  new InstructionFactory(cpg);
	      for(int i = 0; i < methods.length; i++){
	         Method currMeth = methods[i];
	         MethodGen currGen = new MethodGen(currMeth,clazz.getClassName(), cpg);
			 InstructionList il = currGen.getInstructionList();
	         if(il != null){
	           for(InstructionHandle ih = il.getStart();ih != null ; ih = ih.getNext()){
	               Instruction inst= ih.getInstruction();
	               if(inst instanceof INVOKESTATIC){
	                   INVOKESTATIC invoke =(INVOKESTATIC)inst;
	                   Iterator mergeSetItr = myMergeSets.values().iterator();
	                   while(mergeSetItr.hasNext()){
	                   		MergeInfo info = ((MergeInfo)mergeSetItr.next());
	                        ArrayList mergeSet = info.mergeSet;
	                        String className = info.className;
	                        for(int p = 0; p < mergeSet.size(); p++){
	                            Method possible =(Method)mergeSet.get(p);
	                            if(invoke.getClassName(cpg).equals(info.className) &&
	                            	invoke.getMethodName(cpg).equals(possible.getName()) &&
	                                invoke.getSignature(cpg).equals(possible.getSignature())){
	                                	Instruction newinst = new BIPUSH((byte)p);
	                                	ih.setInstruction(newinst);
	                                	inst = factory.createInvoke(info.className,info.methodName,
	                                                            info.retType,
	                                                            info.argTypes,
	                                                            Constants.INVOKESTATIC);
	                                	il.append(ih, inst);
	                                	ih = ih.getNext();
	                		     	}
	                			}
	             			}
	          			}
	        		}
	       		}
	            currGen.setInstructionList(il);
	            currGen.setMaxStack();
	            currGen.setMaxLocals();
	           	cgen.replaceMethod(currMeth, currGen.getMethod());
	        }
	        JavaClass newClass =  cgen.getJavaClass();
	        ConstantPool pool =   cpg.getFinalConstantPool();
	        newClass.setConstantPool(pool);
	        newClass.dump(new java.io.File("C:/JHide/Obfuscated_Classes/",newClass.getClassName().replace('.', java.io.File.separatorChar) + "_obf" +".class"));
	        convertToHTML(newClass,"Obf");


    }

    public static void convertToHTML(JavaClass clazz,String type)   throws java.io.IOException {
         Class2HTML conObj = new Class2HTML(clazz,"C:/JHide/Obfuscated_Classes/HTMLFiles/" + type);
    }
}
