package algorithms;

/**
@ Class: NameOverloading
@ Author: Suma Venkatesh
@ Description:   This algorithm is based on Paul Tyma's patented algorithm.
  This algorithm obfuscates methods so that as many methods as possible have
  the same name
*/




public class NameOverloading {
	    /**
	     * suffix used to create names of the format "_" + my_suffix
	     */
	    private int my_suffix;
	
	    /**
	     * the TreeSet that represents the method node graph
	     */
	    private java.util.TreeSet my_graph;
	
	    /**
	     * set of the new names currently assigned to the methods
	     */
	    private java.util.TreeSet my_names;
	
	    /**
	     * map of the old names to corresponding new names
	     */
	    private java.util.HashMap my_renameLookup;
	
	    /**
	     * all of the indexes that have changed in the constant pool(s)
	     */
	    private java.util.LinkedList my_changedIndicies;
	
	    /**
	     * inheritance graph used to determine which methods can be
	     * renamed.
	     */
	    private utils.ClassHierarchy my_classHierarchy;

   

	    /**
	     * creates an obfuscator
	     * @param label label used to identify obfuscation in gui
	     */
	    public NameOverloading() {
	        my_suffix = 0;
	        my_names = new java.util.TreeSet();
	        my_renameLookup = new java.util.HashMap();
	        my_changedIndicies = new java.util.LinkedList();
	
	    }

	    /**
	     * applies the given obfuscation to the ClassFileCollection
	     * passed as an argument (overrides abstract method defined
	     * in superclass, AppObfuscator)
	     * @param cfc ClassFileCollection contianing the applications
	     * classfiles
	     */
	    public void apply(utils.ClassFileCollection cfc)  {
	        try {
	            // get a iterator over the classes in the ClassFileCollection
	            java.util.Iterator i = cfc.classes();
	            java.util.LinkedList tmp = new java.util.LinkedList();
	
	            // while there are classes not visited
	            while(i.hasNext()){
	                // get the string representation of the file
	                String file = (String)i.next();
	                tmp.add(file);
	            }
	
	
	            Object[] cTmp = tmp.toArray();
	
	           
	
	            String[] classes = new String[cTmp.length];
	            for(int j=0; j<cTmp.length; j++)
	                classes[j] = (String)cTmp[j];
	
	            // initialize the the ClassHierarchy object for this particular
	            // application.  Since each hierarchy is different for each
	            // application - the initialize happens in apply.
	            java.util.Properties props = new java.util.Properties();
	            my_classHierarchy = new utils.ClassHierarchy(cfc, props);
	
	            // class overloaded apply which performs the transformations
	            overload_names(classes, cfc);
	
	
	        } catch(java.io.IOException ioe) {
	            
	            System.out.println("**ERROR** Cannot locate or open Jar file\n");
	            ioe.printStackTrace();
	        }
	    }


    /**
     * applies the given obfuscation to the array of JavaClass objects is
     * passed as parameters.
     * @param classes JavaClass objects that make up the application being
     * obfuscated.
     */
    public void overload_names(String[] classes,utils.ClassFileCollection cfc)
    throws java.io.IOException
    {
        my_graph = collectMethodNodes(classes, cfc);
        System.out.println("After collectMethodNodes");
        buildGraph(my_graph);
        System.out.println("After buildGraph");
        
        System.out.println("After colorGraph");
        makeLookup();
        System.out.println("After makeLookup");
        commitNameChanges(classes, cfc);
        System.out.println("After commitNameChanges");
    }

    /**
     * performs that actual changes of the methods names
     * @param classes JavaClass objects that make up the application being
     * obfuscated
     */
    protected boolean commitNameChanges(String[] classes,
                                        utils.ClassFileCollection cfc)
    {
        
        de.fub.bytecode.classfile.JavaClass jClass = null;

        // do for all classes...
        for(int i =0; i < classes.length; i++)
        {   // dump changed indicies out... will fill this with new "changes" upon each iteration
                my_changedIndicies.clear();

                try {

                    jClass = getClass(classes[i]);
                } catch(java.io.IOException notInJar) {
                    System.out.println("\tfile not contained in Jar\n" + notInJar);
                    return false;
                }

                String className = jClass.getClassName();

                changeOwnNames(jClass);
                changeOtherNames(jClass);
                try{
                    addClass(jClass);
                    convertToHTML(jClass,"Obf");
                }catch(java.io.IOException ioe){
                    System.out.println("\tError writing to class file collection...aborting.");
                    return false;
                }

        }
        return true;
    }

    public static void convertToHTML(de.fub.bytecode.classfile.JavaClass clazz,String type)   throws java.io.IOException {
	   de.fub.bytecode.util.Class2HTML conObj = new de.fub.bytecode.util.Class2HTML(clazz,"C:/JHide/Obfuscated_Classes/HTMLFiles/" + type);
	}

    /**
	 *  Adds or replaces a class in this collection, given a JavaClass.
	 */
	public void addClass(de.fub.bytecode.classfile.JavaClass c) throws java.io.IOException {

			c.dump(new java.io.File("C:/JHide/Obfuscated_Classes/",c.getClassName().replace('.', java.io.File.separatorChar) + "_obf" +".class"));

	}

    // change names of members which are defined in this class
    private void changeOwnNames(de.fub.bytecode.classfile.JavaClass jClass)
    {
        de.fub.bytecode.classfile.Field[] theFields = jClass.getFields();
        de.fub.bytecode.classfile.Method[] theMethods = jClass.getMethods();
        de.fub.bytecode.classfile.ConstantPool pool = jClass.getConstantPool();

        //makeChanges(theFields, pool, false);
        makeChanges(theMethods, pool, true);
    }

    // refactored method - came from changeOwnNames
    private void makeChanges(de.fub.bytecode.classfile.FieldOrMethod[] targets,
                             de.fub.bytecode.classfile.ConstantPool cPool,
                             boolean methVersion)
    {
        int index;
        de.fub.bytecode.classfile.ConstantUtf8 utf8New;
        String temp, newName;

        for( int i=0 ; i < targets.length ; i++)
        {
            temp = targets[i].getName();

            // if we are processing Methods, and the name of the temp is
            // equal to a name in the Java SDK that we cannot/should not
            // change - then return to the top of the loop and do not
            // execute the code below this point
            if( methVersion && isNotChangeable(temp) )
                continue;

            // if we reach this point we are allowed to make changes
            // to the targets
            newName = (String)my_renameLookup.get(temp);
            System.out.println(my_renameLookup);
            if( newName != null )
            {

            utf8New = new de.fub.bytecode.classfile.ConstantUtf8(newName);
            index = targets[i].getNameIndex();
            cPool.setConstant(index,utf8New);

            my_changedIndicies.add(new Integer(index));
            }
        }
    }

    // change names of external references
    private void changeOtherNames(de.fub.bytecode.classfile.JavaClass jClass)
    {
        de.fub.bytecode.classfile.ConstantPool pool = jClass.getConstantPool();
        de.fub.bytecode.classfile.Constant[] cArray = pool.getConstantPool();

       
        String owner;
        for(int i = 0; i< cArray.length; i++)
        {
            if( cArray[i] instanceof de.fub.bytecode.classfile.ConstantMethodref ||
                cArray[i] instanceof de.fub.bytecode.classfile.ConstantFieldref )
            {
                int classIndex = ((de.fub.bytecode.classfile.ConstantCP)cArray[i]).getClassIndex();
                owner = pool.constantToString(pool.getConstant(classIndex));

                // if the reference is defined in the current class, continue
                if( !owner.equals(jClass.getClassName()) )
                {
                    // get the string to be replaced in oldname

                    String oldname = pool.constantToString(cArray[i]);
                    java.util.StringTokenizer tok = new java.util.StringTokenizer(oldname,".");

                    String pkgName = tok.nextToken();

                    if( !(pkgName.equals("java") ||
                          pkgName.equals("javax")) )
                    {

                        while(tok.hasMoreTokens())
                        {
                            oldname = tok.nextToken();
                        }

                        tok = new java.util.StringTokenizer(oldname," ");
                        oldname = tok.nextToken();

                        // dont change constructors
                        if( !oldname.equals("<init>") )
                        {
                            String newname = "";

                            if(my_renameLookup.containsKey(oldname))
                            {
                                newname = (String)my_renameLookup.get(oldname);
                            }
                            else
                            {
                                System.out.println("\t>>>why is there no name???");
                                
                            }

                            for (int j=0 ; j < cArray.length ; j++)
                            {
                                if(cArray[j] instanceof de.fub.bytecode.classfile.ConstantUtf8)
                                {
                                    if (((de.fub.bytecode.classfile.ConstantUtf8)cArray[j]).getBytes().compareTo(oldname) == 0)
                                    {
                                        // if this constant has been changed earlier, dont change it again

                                        if( !my_changedIndicies.contains(new Integer(j)))
                                            ((de.fub.bytecode.classfile.ConstantUtf8)cArray[j]).setBytes(newname);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
             
    } // end of method

    /**
     * creates a mapping of old method name to new method name that can be
     * changed when the final changes are being made to the classes
     */
    private void makeLookup()
    {
        java.util.Iterator itr = my_graph.iterator();

        while(itr.hasNext())
        {
            utils.MethodNode node = (utils.MethodNode)itr.next();

            my_renameLookup.put(node.getMethodName(), node.getNewName());
        }
    }

   

    /**
     * collects all the methods of the classes in the application.  A MethodNode object is
     * created for each method that stores information about the class if came from, its
     * original name, and its signature (or parameter list).  If the method is associated
     * in any way with the Java 2 SDK, it is not added to the method node graph because
     * the transformation being applied cannot modify its method name.
     *
     * @param classes an array of JavaClass objects that make up the application being
     * obfuscated.
     */
    private java.util.TreeSet collectMethodNodes(String[] classes,
                                                 utils.ClassFileCollection cfc)
                                                 throws java.io.IOException
    {
        de.fub.bytecode.classfile.JavaClass jClass = null;

        java.util.TreeSet methNodes = new java.util.TreeSet();

        // do for all classes...
        for( int i =0; i < classes.length; i++ ){

            //jClass = cfc.getClass(classes[i]);

			jClass = getClass(classes[i]);

            String className = jClass.getClassName();

            de.fub.bytecode.classfile.Method[] theMethods = jClass.getMethods();
            for(int m = 0; m < theMethods.length; m++)
            {
                String mName = theMethods[m].getName();
                String mSig = parseSignature(theMethods[m].getSignature());

                // create a tmp MethodNode
                utils.MethodNode tmp =
                    new utils.MethodNode(className, mName, mSig);

                // create an identical MethodNode with a different name
                //utils.MethodNode possNew =
                //    new utils.MethodNode(className, mName+"__", mSig);

                utils.MethodID oldN = new utils.MethodID(mName, mSig, className);

                // query the ClassHierarchy object to determine if it is ok
                // for this method to be renamed.  If this method is part of
                // a Java 2 SDK interface or is overriding behavior defined
                // by a Java 2 SDK class at some point in the inheritance
                // hierarchy - then the obfuscating transformation cannot
                // method the method name... therefore do not add to graph

                System.out.println("old:"+oldN);

                try {
                 
                       methNodes.add(tmp);
                }catch(Exception exc) {
                   System.out.println("Exception is:::: " + exc);
                }
            }
        }

        return methNodes;
    }

    public de.fub.bytecode.classfile.JavaClass getClass(String classname) throws java.io.IOException {
		//java.io.File f = new java.io.File("C:/JHide/examples/TestNameOverloading.class");
		java.io.File f = new java.io.File("C:/JHide/examples/" + classname +".class");
	    de.fub.bytecode.classfile.ClassParser p = new de.fub.bytecode.classfile.ClassParser(f.getPath());
	    return p.parse();
}

    /**
     * method used to get the next method name, the method names are in the format
     * underscore and suffix, "_#"
     * @return String a String representing the next valid method name
     */
    private String getNextName()
    {
        return "_" + my_suffix++;
    }

    /**
     * determines is the String passed as a parameter represents a "special" method
     * that cannot be changed.  Examples of such methods are main(String[]), toString(),
     * and all constructors (denoted as <init> in the classfile format).
     * @return boolean true is the String passed is for a method that we cannot change,
     * else false
     */
    private boolean isNotChangeable(String methodName)
    {
        return (methodName.equals("<init>") ||
                methodName.equals("main")   ||
                methodName.equals("toString"));
    }

    /**
     * connects the nodes contained in the LinkedList structure passed
     * as a parameter.
     * @param list a list of nodes that are to be connected, e.g. edges
     * will be created between all of them.  Any nodes in this list
     * will make up a subgraph of the whole graph that is a "complete"
     * graph, meaning that for each node in the subgraph, there exists an
     * edge from it to all the other nodes in the subgraph.
     */
    private void connectNodes(java.util.LinkedList list)
    {
        java.util.Iterator i = list.iterator();

        while(i.hasNext())
        {
            utils.MethodNode tmp = (utils.MethodNode)i.next();

            for(int n = 0; n < list.size(); n++)
            {
                utils.MethodNode node = (utils.MethodNode)list.get(n);
                tmp.addEdge(node);
            }
        }
    }

    /**
     * take a TreeSet object that contains MethodNode object, these object make up the
     * method node graph.  This method creates the vital edges between nodes that
     * have the same signature.  Thus, if a node exists between two nodes, s and t,
     * then s and t cannot have the same name assigned to them by the transformation.
     *
     * @param set a TreeSet containing MethodNode object that make up the graph
     */
    private void buildGraph(java.util.TreeSet set)
    {
        // ###PROGRAMMERS NOTE###
        // each TreeSet has the property that it is maintained in
        // sorted order.  Therefore, it is not like the traditional definition
        // or a set, which is an unordered collection of elements of the same type
        // ###PLEASE NOTE THIS###
        //     - the following code takes advantage of this sorted property and
        //       would not function correctly if this structure was not used.
        System.out.println(set);
        // examine first node
        utils.MethodNode firstNode = (utils.MethodNode)set.first();
        // set current signature to the first node's signature
        String currSig = firstNode.getMethodSignature();


        java.util.Iterator i = set.iterator();
        // temp structure to store refs to node before
        // creating edges between them
        java.util.LinkedList list = new java.util.LinkedList();

        while(i.hasNext())
        {
            utils.MethodNode tmp = (utils.MethodNode)i.next();
            String tmpSig = tmp.getMethodSignature();

            // the TreeSet could look something like this...
            // [(byte), (double), (double), double), (int), (int), (int),
            //  (Object), (Object), (Object), (String)]
            // therefore, if the current signature is the same as the
            // tmp signature - add it to the temporary storage structure.
            // what this will do is give us a temp structure that contains
            // signatures of one type - these are the node we wish to create
            // edges between.

            if(currSig.equals(tmpSig))
                list.add(tmp);
            else
            {   // if we reach this, we are looking at a different signature
                // group, so we want to connect all the node that have been
                // added to the temp structure, clear that storage so new
                // nodes can be placed in it... and set the current signature
                // to the tmp signature.
                currSig = tmp.getMethodSignature();
                connectNodes(list);
                list.clear();
                list.add(tmp);
            }
        }
        // make the connections between the nodes in the list
        // when we exit the loop, we ill most likely have nodes left in
        // the temporary structure... so we need to make the proper connection,
        // edges, between them. if we don't, the transformation might assign
        // them the same name (bad, since they share method signutures).
        connectNodes(list);
    }

    /**
     * parses the String passed as a method, the String represents the signature
     * of a method.  The format of the String is
     * (parameter, parameter, ...) return type.  Return types to do not uniquely
     * identify methods, so this information needs to be stripped off.
     * @param sig String representing a method signature
     * @return String a String presenting the signature of a method with the
     * return type stripped off.
     */
    public static String parseSignature(String sig)
    {
        return sig.substring(0,sig.indexOf(")")+1);
    }

   

    public boolean apply(String[] andy)   {
	        NameOverloading no =  new NameOverloading();
	        boolean sucess = false;

	       String[] temp = new String[andy.length-1];
	       System.arraycopy(andy, 1, temp, 0, temp.length);
	       try {
	              no.overload_names(temp, new utils.ClassFileCollection());
	              sucess = true;
	       } catch(java.io.IOException ioe) {
	         System.out.println("***error - file obfuscation not attempted***" +ioe);
	       }	      

	      
	        return sucess;
	 }

}












