
package algorithms;
import org.apache.bcel.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.Class2HTML;

/**
@ Class: SplittingVariables
@ Author: Suma Venkatesh
@ Description:   This algorithm is a kind of splitting transformation where types of the variables are split into smaller ones.
**/
public class SplittingVariables {

	public boolean apply(String classname) {
   	   boolean sucess = false;
		try {
			JavaClass clazz = Repository.lookupClass("examples/" + classname);
			SplittingVariables obj=new SplittingVariables();
			obj.variableIdentifier(clazz);
			obj.init(clazz);
			sucess = true;
		 }
		catch(Exception e) {
			System.out.println("Exception " +e.fillInStackTrace()); return(sucess);
		}
		return(sucess);
	}

	public void init(JavaClass theClass) {

		int i,j,slot,memloc, argsize, newloc;
		try{
			String classname = theClass.getClassName();
		 	ClassGen cgen = new ClassGen(theClass);
			Method[] methods = theClass.getMethods();
 		    MethodGen mymgen = null;
		 	ConstantPoolGen cpg = new ConstantPoolGen(theClass.getConstantPool());
			slot=0;
			 // Find the method to modify
			 for (i=0; i < methods.length; i++) {
					mymgen = new MethodGen(methods[i], classname, cpg);
			   		slot=i;
			 }

		 	  if (mymgen == null) throw new IllegalArgumentException("No such method to obfuscate.");
					 if ((mymgen.isInterface() || mymgen.isAbstract())) {
					 // Dodnt do
					 return;
		 	  }
		  	  InstructionList il= mymgen.getInstructionList();
		 	  LocalVariableGen[] localvargens= mymgen.getLocalVariables();
			  InstructionHandle[] ihs=il.getInstructionHandles();
	 		  int thissize;
		 	  argsize=1;
		 	  // Find the argument size...
		 	  for (i=0; i<localvargens.length; i++) {
					thissize=localvargens[i].getIndex();
					thissize+=localvargens[i].getType().getSize();
					if (thissize>argsize) {
			   			argsize=thissize;
					}
		 		}
			 int local_slots[] = new int[ihs.length*4+localvargens.length*2];
			 int local_inst[] = new int[ihs.length];
			 newloc=0;
		 	 // The local variables are the arguments of the method
		 	 for (j=0; j<ihs.length; j++) {
				// If this is a store/load/inc instruction... that doesn't refer to the arguments
				if ((ihs[j].getInstruction() instanceof StoreInstruction)&&(((LocalVariableInstruction) ihs[j].getInstruction()).getIndex()>=argsize)) {
			   		int index=((LocalVariableInstruction) ihs[j].getInstruction()).getIndex();
			   		if (((ihs[j].getInstruction() instanceof ASTORE))||((ihs[j].getInstruction() instanceof DSTORE))||((ihs[j].getInstruction() instanceof LSTORE))) {
				  		// Don't use this one...
				  		local_slots[index]=2;
			   		}
					else {
			  	    	if (local_slots[index]==0) {
							 local_slots[index]=1;
					 	 	 local_inst[index]=j;
				  	    }
			    	}
				}
		     }

			 // randomly choosing a localvariable to use...
			 for (i=0,memloc=0; i<local_slots.length-1; i++) {
				if (local_slots[i]==1) {
					 local_slots[memloc++]=i;
			   		 local_inst[memloc]=local_inst[i];
				}
		 	 }
			 if (memloc==0) {
				// Nothing to do
				return;
		 	  }
	 		  memloc = local_slots[((int) (((int) (Math.random()*memloc))/memloc))];
	 		  LocalVariableGen newvar = mymgen.addLocalVariable("",((LocalVariableInstruction) ihs[local_inst[memloc]].getInstruction()).getType(cpg),null,null);

		  	  newloc=newvar.getIndex();

		 	  //call the actual splitting method
	 		  split_vars(il, memloc, newloc);
			  // Put instructions back
		 	  mymgen.setInstructionList(il);
		 	  mymgen.setMaxStack();
		 	  methods[slot]=mymgen.getMethod();
		 	  cgen.setMethods(methods);
		 	  JavaClass jc = cgen.getJavaClass();

		 	  //commit changes and dump in obfuscated classes
		 	  jc.setConstantPool(cpg.getFinalConstantPool());
			  jc.dump(new java.io.File("C:/JHide/Obfuscated_Classes/",jc.getClassName().replace('.', java.io.File.separatorChar) + "_obf"+ ".class"));
			  convertToHTML(jc,"Obf");

  	    }
		catch(Exception e) {System.out.println("Exception" +e.getMessage() + e.fillInStackTrace()); }
	}




	private void split_vars(InstructionList instlist, int old_slot, int new_slot) {

		 InstructionHandle curih;

		  for (curih=instlist.getStart(); curih!=instlist.getEnd(); curih=curih.getNext()) {

			 // Is this a load instruction for the old_slot?
			 if ((curih.getInstruction() instanceof LoadInstruction)&&(((LocalVariableInstruction) curih.getInstruction()).getIndex()==old_slot)) {
				if (Math.random()<.5) {
				   ((LocalVariableInstruction) curih.getInstruction()).setIndex(new_slot);
				}
			 }

			 // Is this a store instruction for the old_slot?
			 if ((curih.getInstruction() instanceof StoreInstruction)&&(((LocalVariableInstruction) curih.getInstruction()).getIndex()==old_slot)) {
				// Append the original instruction to itself...
				instlist.append(curih,curih.getInstruction().copy());
				((LocalVariableInstruction) curih.getInstruction()).setIndex(new_slot);
				// Append xSTORE new_slot to the original instruction...
				instlist.append(curih,curih.getInstruction().copy());
				curih.setInstruction(new DUP());

				// Move the curih location past the new store inst...
				curih=curih.getNext().getNext();
			 }
			 // Is this INC for the old_slot?
			 if ((curih.getInstruction() instanceof IINC)&&(((LocalVariableInstruction) curih.getInstruction()).getIndex()==old_slot)) {
				// Append the instruction to itself to the original instruction...
				instlist.append(curih,curih.getInstruction().copy());
				((LocalVariableInstruction) curih.getInstruction()).setIndex(new_slot);

				// Move the curih location past the inc insts...
				curih=curih.getNext();
			 }

		  }
		  instlist.setPositions();
	   }

	  public void variableIdentifier(JavaClass homeClass)  throws java.io.IOException {


			 ClassGen cgen = new ClassGen(homeClass);
			 Method[] methods = homeClass.getMethods();
			 int m = 0;
			 MethodGen mgen = null;
			 ConstantPoolGen cpg = new ConstantPoolGen(homeClass.getConstantPool());
			 String classname = homeClass.getClassName();
			 for(; m < methods.length; m++){
				   mgen = new MethodGen(methods[m], classname, cpg);
				   if(mgen == null) throw new IllegalArgumentException("No such method  to obfuscate.");
				   if(mgen.isInterface() || mgen.isAbstract()) return;
				   cgen.replaceMethod(methods[m], mgen.getMethod());
				   cgen.setConstantPool(cpg);
			 }
	  }
	  public static void convertToHTML(JavaClass clazz,String type)   throws java.io.IOException {
   		     Class2HTML conObj = new Class2HTML(clazz,"C:/JHide/Obfuscated_Classes/HTMLFiles/" + type);
      }
}
