package engine;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;

import algorithms.AlterIfElse;
import algorithms.ClassSplitter;
import algorithms.ConstantsPoolReordering;
import algorithms.DataAggregation;
import algorithms.DataEncoding;
import algorithms.DataOrdering;
import algorithms.HighLevelLanguageBreaking;
import algorithms.LocalVariableReordering;
import algorithms.LoopBlocking;
import algorithms.LoopUnroll;
import algorithms.MethodCloning;
import algorithms.NameOverloading;
import algorithms.NewAlgorithm;
import algorithms.NodeSplitter;
import algorithms.ScramblingIdentifier;
import algorithms.FormatRemoval;
import algorithms.InsertDeadCodeFields;
import algorithms.InsertDeadCodeLoops;
import algorithms.InsertDeadCodeMethods;
import algorithms.InsertDeadCodeMethodArguments;
import algorithms.MethodInLiner;
import algorithms.ParameterReordering;
import algorithms.SplittingVariables;
import algorithms.DataStorage;
import algorithms.MethodsMerging;
import algorithms.ParameterMerging;
import algorithms.ClassesMerging;
import algorithms.LoopFission;
import algorithms.LoopReorder;
import algorithms.Method2RMadness.Method2RMadness;


import org.apache.bcel.util.Class2HTML;

/**
@ Class: ApplyTransformations
@ Author: Suma Venkatesh
@ Description:   This class is the main engine which calls the repective algorithms based on the input from ObfuscationView.
**/

public class ApplyTransformations {

	/* This method is called by OfuscationView and it delegated the control to individual algorithms.
	   Takes the input class name and algorithm name as the parameter.
	*/

	public boolean callEngine(String classname,String algname) throws java.io.IOException{
	   boolean sucess = false;
	   try {

		   int i = classname.indexOf(".");
		   classname = classname.substring(0,i);


		   JavaClass clazz = Repository.lookupClass("examples/" +  classname.trim());

		   ClassGen cgen = new ClassGen(clazz);
		   ConstantPoolGen cpg = cgen.getConstantPool();

		   if(algname.equalsIgnoreCase("FormatRemoval")){
				convertToHTML(clazz,"Src");
				FormatRemoval obj = new FormatRemoval();
				sucess = obj.apply(classname);
		   }

		   if(algname.equalsIgnoreCase("ScramblingIdentifier")){
		   		convertToHTML(clazz,"Src");
		   		ScramblingIdentifier obj = new ScramblingIdentifier();
		   		sucess = obj.apply(clazz);
		   }
		   if(algname.equalsIgnoreCase("InsertDeadCodeLoops")){
		   		convertToHTML(clazz,"Src");
		   		InsertDeadCodeLoops obj = new InsertDeadCodeLoops();
		   		sucess = obj.apply(clazz,classname);
		   }
 		   if(algname.equalsIgnoreCase("InsertDeadCodeMethodArguments")){
			   	convertToHTML(clazz,"Src");
				InsertDeadCodeMethodArguments obj = new InsertDeadCodeMethodArguments();
				sucess = obj.apply(classname);
		   }
		   if(algname.equalsIgnoreCase("InsertDeadCodeFields")){
				convertToHTML(clazz,"Src");
				InsertDeadCodeFields obj = new InsertDeadCodeFields();
				sucess = obj.apply(classname);
		   }

			if(algname.equalsIgnoreCase("InsertDeadCodeMethods")){

				convertToHTML(clazz,"Src");
				InsertDeadCodeMethods obj = new InsertDeadCodeMethods();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("LoopFission")){
				convertToHTML(clazz,"Src");
				LoopFission obj = new LoopFission();
				sucess = obj.apply(classname);
			}
            if(algname.equalsIgnoreCase("DataStorage")){
				convertToHTML(clazz,"Src");
				DataStorage obj = new DataStorage();
				sucess = obj.apply(classname);
			}

			if(algname.equalsIgnoreCase("LoopReorder")){
				convertToHTML(clazz,"Src");
				LoopReorder obj = new LoopReorder();
				sucess = obj.apply(classname);
				sucess = true;
			}

			if(algname.equalsIgnoreCase("SplittingVariables")){
				convertToHTML(clazz,"Src");
				SplittingVariables obj = new SplittingVariables();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("MethodsMerging")){
				convertToHTML(clazz,"Src");
				MethodsMerging obj = new MethodsMerging();
				sucess =obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("ParameterMerging")){
				convertToHTML(clazz,"Src");
				ParameterMerging obj = new ParameterMerging();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("ClassesMerging")){

				convertToHTML(clazz,"Src");

				ClassesMerging obj = new ClassesMerging();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("ParameterReordering")){
				convertToHTML(clazz,"Src");
				ParameterReordering obj = new ParameterReordering();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("MethodInLiner")){
				convertToHTML(clazz,"Src");
				MethodInLiner obj = new MethodInLiner();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("LocalVariableReordering")){
				convertToHTML(clazz,"Src");
				LocalVariableReordering obj = new LocalVariableReordering();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("ConstantsPoolReordering")){
				convertToHTML(clazz,"Src");
				ConstantsPoolReordering obj = new ConstantsPoolReordering();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("LoopUnroll")){
				convertToHTML(clazz,"Src");
				LoopUnroll obj = new LoopUnroll();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("AlterIfElse")){
				convertToHTML(clazz,"Src");
				AlterIfElse obj = new AlterIfElse();
				sucess = obj.apply(clazz);
			
			}
			if(algname.equalsIgnoreCase("HighLevelLanguageBreaking")){
				convertToHTML(clazz,"Src");
				HighLevelLanguageBreaking obj = new HighLevelLanguageBreaking();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("NodeSplitter")){
				convertToHTML(clazz,"Src");
				NodeSplitter obj = new NodeSplitter();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("DataAggregation")){
				convertToHTML(clazz,"Src");
				DataAggregation obj = new DataAggregation();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("ClassSplitter")){
				convertToHTML(clazz,"Src");
				ClassSplitter obj = new ClassSplitter();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("NameOverloading")){

				convertToHTML(clazz,"Src");
				NameOverloading obj = new NameOverloading();
				String[] param = new String[2];
				param[0] = "-f";
				param[1] = classname;
				sucess = obj.apply(param);

			}
			if(algname.equalsIgnoreCase("Method2RMadness")){
				convertToHTML(clazz,"Src");
				Method2RMadness obj = new Method2RMadness();
				String[] param = new String[1];
				param[0] = classname;
				sucess = obj.driver(param);
			}
			if(algname.equalsIgnoreCase("MethodCloning")){
				convertToHTML(clazz,"Src");
				MethodCloning obj = new MethodCloning();
				sucess = obj.apply(classname);
			}

			if(algname.equalsIgnoreCase("LoopBlocking")){
				convertToHTML(clazz,"Src");
				LoopBlocking obj = new LoopBlocking();
				sucess = obj.apply(classname);
			}
			if(algname.equalsIgnoreCase("DataEncoding")){
				convertToHTML(clazz,"Src");
				DataEncoding obj = new DataEncoding();
				sucess = obj.apply(clazz);
			}
			if(algname.equalsIgnoreCase("DataOrdering")){
				convertToHTML(clazz,"Src");
				DataOrdering obj = new DataOrdering();
				sucess = obj.apply(clazz);
			}
			if(algname.equalsIgnoreCase("NewAlgorithm")){
				convertToHTML(clazz,"Src");
				NewAlgorithm obj = new NewAlgorithm();
				obj.test();
			}
	   } catch(Exception e) {System.out.println("Exception is : " + e);}
	   return(sucess);

    }

    /* This method converts an input class file into set of byte code instructions and dumps it into a HTML file
    */
    public static void convertToHTML(JavaClass clazz,String type)   throws java.io.IOException {
		   Class2HTML conObj = new Class2HTML(clazz,"C:/JHide/Obfuscated_Classes/HTMLFiles/" + type);
    }
}